@@ -1,8 +1,8 @@
+from __future__ import annotations
+
 import os.path
 
-pytest_plugins = [
-    'mypy.test.data',
-]
+pytest_plugins = ["mypy.test.data"]
 
 
 def pytest_configure(config):
@@ -14,5 +14,6 @@ def pytest_configure(config):
 # This function name is special to pytest.  See
 # http://doc.pytest.org/en/latest/writing_plugins.html#initialization-command-line-and-configuration-hooks
 def pytest_addoption(parser) -> None:
-    parser.addoption('--bench', action='store_true', default=False,
-                     help='Enable the benchmark test runs')
+    parser.addoption(
+        "--bench", action="store_true", default=False, help="Enable the benchmark test runs"
+    )
