@@ -3,21 +3,17 @@
 These can be used for filtering specific errors.
 """
 
-from typing import Dict, List
+from __future__ import annotations
+
 from typing_extensions import Final
 
-
-# All created error codes are implicitly stored in this list.
-all_error_codes: List["ErrorCode"] = []
-
-error_codes: Dict[str, "ErrorCode"] = {}
+error_codes: dict[str, ErrorCode] = {}
 
 
 class ErrorCode:
-    def __init__(self, code: str,
-                 description: str,
-                 category: str,
-                 default_enabled: bool = True) -> None:
+    def __init__(
+        self, code: str, description: str, category: str, default_enabled: bool = True
+    ) -> None:
         self.code = code
         self.description = description
         self.category = category
@@ -25,7 +21,7 @@ class ErrorCode:
         error_codes[code] = self
 
     def __str__(self) -> str:
-        return '<ErrorCode {}>'.format(self.code)
+        return f"<ErrorCode {self.code}>"
 
 
 ATTR_DEFINED: Final = ErrorCode("attr-defined", "Check that attribute exists", "General")
@@ -37,7 +33,9 @@ ARG_TYPE: Final = ErrorCode("arg-type", "Check argument types in calls", "Genera
 CALL_OVERLOAD: Final = ErrorCode(
     "call-overload", "Check that an overload variant matches arguments", "General"
 )
-VALID_TYPE: Final = ErrorCode("valid-type", "Check that type (annotation) is valid", "General")
+VALID_TYPE: Final[ErrorCode] = ErrorCode(
+    "valid-type", "Check that type (annotation) is valid", "General"
+)
 VAR_ANNOTATED: Final = ErrorCode(
     "var-annotated", "Require variable annotation if type can't be inferred", "General"
 )
@@ -50,7 +48,7 @@ RETURN: Final[ErrorCode] = ErrorCode(
 RETURN_VALUE: Final[ErrorCode] = ErrorCode(
     "return-value", "Check that return value is compatible with signature", "General"
 )
-ASSIGNMENT: Final = ErrorCode(
+ASSIGNMENT: Final[ErrorCode] = ErrorCode(
     "assignment", "Check that assigned value is compatible with target", "General"
 )
 TYPE_ARG: Final = ErrorCode("type-arg", "Check that generic type arguments are present", "General")
@@ -82,6 +80,9 @@ FUNC_RETURNS_VALUE: Final = ErrorCode(
 ABSTRACT: Final = ErrorCode(
     "abstract", "Prevent instantiation of classes with abstract attributes", "General"
 )
+TYPE_ABSTRACT: Final = ErrorCode(
+    "type-abstract", "Require only concrete classes where Type[...] is expected", "General"
+)
 VALID_NEWTYPE: Final = ErrorCode(
     "valid-newtype", "Check that argument 2 to NewType is valid", "General"
 )
@@ -94,8 +95,19 @@ STR_BYTES_PY3: Final = ErrorCode(
 EXIT_RETURN: Final = ErrorCode(
     "exit-return", "Warn about too general return type for '__exit__'", "General"
 )
-LITERAL_REQ: Final = ErrorCode(
-    "literal-required", "Check that value is a literal", 'General'
+LITERAL_REQ: Final = ErrorCode("literal-required", "Check that value is a literal", "General")
+UNUSED_COROUTINE: Final = ErrorCode(
+    "unused-coroutine", "Ensure that all coroutines are used", "General"
+)
+# TODO: why do we need the explicit type here? Without it mypyc CI builds fail with
+# mypy/message_registry.py:37: error: Cannot determine type of "EMPTY_BODY"  [has-type]
+EMPTY_BODY: Final[ErrorCode] = ErrorCode(
+    "empty-body",
+    "A dedicated error code to opt out return errors for empty/trivial bodies",
+    "General",
+)
+SAFE_SUPER: Final = ErrorCode(
+    "safe-super", "Warn about calls to abstract methods with empty/trivial bodies", "General"
 )
 
 # These error codes aren't enabled by default.
@@ -110,6 +122,7 @@ NO_UNTYPED_CALL: Final = ErrorCode(
 REDUNDANT_CAST: Final = ErrorCode(
     "redundant-cast", "Check that cast changes type of expression", "General"
 )
+ASSERT_TYPE: Final = ErrorCode("assert-type", "Check that assert_type() call succeeds", "General")
 COMPARISON_OVERLAP: Final = ErrorCode(
     "comparison-overlap", "Check that types in comparisons and 'in' expressions overlap", "General"
 )
@@ -124,6 +137,15 @@ NO_ANY_RETURN: Final = ErrorCode(
 UNREACHABLE: Final = ErrorCode(
     "unreachable", "Warn about unreachable statements or expressions", "General"
 )
+ANNOTATION_UNCHECKED = ErrorCode(
+    "annotation-unchecked", "Notify about type annotations in unchecked functions", "General"
+)
+POSSIBLY_UNDEFINED: Final[ErrorCode] = ErrorCode(
+    "possibly-undefined",
+    "Warn about variables that are defined only in some execution paths",
+    "General",
+    default_enabled=False,
+)
 REDUNDANT_EXPR: Final = ErrorCode(
     "redundant-expr", "Warn about redundant expressions", "General", default_enabled=False
 )
@@ -133,6 +155,17 @@ TRUTHY_BOOL: Final[ErrorCode] = ErrorCode(
     "General",
     default_enabled=False,
 )
+TRUTHY_FUNCTION: Final[ErrorCode] = ErrorCode(
+    "truthy-function",
+    "Warn about function that always evaluate to true in boolean contexts",
+    "General",
+)
+TRUTHY_ITERABLE: Final[ErrorCode] = ErrorCode(
+    "truthy-iterable",
+    "Warn about Iterable expressions that could always evaluate to true in boolean contexts",
+    "General",
+    default_enabled=False,
+)
 NAME_MATCH: Final = ErrorCode(
     "name-match", "Check that type definition has consistent naming", "General"
 )
@@ -147,6 +180,21 @@ IGNORE_WITHOUT_CODE: Final = ErrorCode(
     "General",
     default_enabled=False,
 )
+UNUSED_AWAITABLE: Final = ErrorCode(
+    "unused-awaitable",
+    "Ensure that all awaitable values are used",
+    "General",
+    default_enabled=False,
+)
+REDUNDANT_SELF_TYPE = ErrorCode(
+    "redundant-self",
+    "Warn about redundant Self type annotations on method first argument",
+    "General",
+    default_enabled=False,
+)
+USED_BEFORE_DEF: Final[ErrorCode] = ErrorCode(
+    "used-before-def", "Warn about variables that are used before they are defined", "General"
+)
 
 
 # Syntax errors are often blocking.
