@@ -1,36 +1,63 @@
 """Fix up various things after deserialization."""
 
-from typing import Any, Dict, Optional
+from __future__ import annotations
+
+from typing import Any
 from typing_extensions import Final
 
+from mypy.lookup import lookup_fully_qualified
 from mypy.nodes import (
-    MypyFile, SymbolTable, TypeInfo, FuncDef, OverloadedFuncDef,
-    Decorator, Var, TypeVarExpr, ClassDef, Block, TypeAlias,
+    Block,
+    ClassDef,
+    Decorator,
+    FuncDef,
+    MypyFile,
+    OverloadedFuncDef,
+    ParamSpecExpr,
+    SymbolTable,
+    TypeAlias,
+    TypeInfo,
+    TypeVarExpr,
+    TypeVarTupleExpr,
+    Var,
 )
 from mypy.types import (
-    CallableType, Instance, Overloaded, TupleType, TypedDictType,
-    TypeVarType, UnboundType, UnionType, TypeVisitor, LiteralType,
-    TypeType, NOT_READY, TypeAliasType, AnyType, TypeOfAny, ParamSpecType,
-    Parameters, UnpackType, TypeVarTupleType
+    NOT_READY,
+    AnyType,
+    CallableType,
+    Instance,
+    LiteralType,
+    Overloaded,
+    Parameters,
+    ParamSpecType,
+    TupleType,
+    TypeAliasType,
+    TypedDictType,
+    TypeOfAny,
+    TypeType,
+    TypeVarTupleType,
+    TypeVarType,
+    TypeVisitor,
+    UnboundType,
+    UnionType,
+    UnpackType,
 )
 from mypy.visitor import NodeVisitor
-from mypy.lookup import lookup_fully_qualified
 
 
 # N.B: we do a allow_missing fixup when fixing up a fine-grained
 # incremental cache load (since there may be cross-refs into deleted
 # modules)
-def fixup_module(tree: MypyFile, modules: Dict[str, MypyFile],
-                 allow_missing: bool) -> None:
+def fixup_module(tree: MypyFile, modules: dict[str, MypyFile], allow_missing: bool) -> None:
     node_fixer = NodeFixer(modules, allow_missing)
     node_fixer.visit_symbol_table(tree.names, tree.fullname)
 
 
 # TODO: Fix up .info when deserializing, i.e. much earlier.
 class NodeFixer(NodeVisitor[None]):
-    current_info: Optional[TypeInfo] = None
+    current_info: TypeInfo | None = None
 
-    def __init__(self, modules: Dict[str, MypyFile], allow_missing: bool) -> None:
+    def __init__(self, modules: dict[str, MypyFile], allow_missing: bool) -> None:
         self.modules = modules
         self.allow_missing = allow_missing
         self.type_fixer = TypeFixer(self.modules, allow_missing)
@@ -48,19 +75,29 @@ class NodeFixer(NodeVisitor[None]):
                 for base in info.bases:
                     base.accept(self.type_fixer)
             if info._promote:
-                info._promote.accept(self.type_fixer)
+                for p in info._promote:
+                    p.accept(self.type_fixer)
             if info.tuple_type:
                 info.tuple_type.accept(self.type_fixer)
+                info.update_tuple_type(info.tuple_type)
+                if info.special_alias:
+                    info.special_alias.alias_tvars = list(info.defn.type_vars)
             if info.typeddict_type:
                 info.typeddict_type.accept(self.type_fixer)
+                info.update_typeddict_type(info.typeddict_type)
+                if info.special_alias:
+                    info.special_alias.alias_tvars = list(info.defn.type_vars)
             if info.declared_metaclass:
                 info.declared_metaclass.accept(self.type_fixer)
             if info.metaclass_type:
                 info.metaclass_type.accept(self.type_fixer)
             if info._mro_refs:
-                info.mro = [lookup_fully_qualified_typeinfo(self.modules, name,
-                                                            allow_missing=self.allow_missing)
-                            for name in info._mro_refs]
+                info.mro = [
+                    lookup_fully_qualified_typeinfo(
+                        self.modules, name, allow_missing=self.allow_missing
+                    )
+                    for name in info._mro_refs
+                ]
                 info._mro_refs = None
         finally:
             self.current_info = save_info
@@ -75,8 +112,9 @@ class NodeFixer(NodeVisitor[None]):
                 if cross_ref in self.modules:
                     value.node = self.modules[cross_ref]
                 else:
-                    stnode = lookup_fully_qualified(cross_ref, self.modules,
-                                                    raise_on_missing=not self.allow_missing)
+                    stnode = lookup_fully_qualified(
+                        cross_ref, self.modules, raise_on_missing=not self.allow_missing
+                    )
                     if stnode is not None:
                         assert stnode.node is not None, (table_fullname + "." + key, cross_ref)
                         value.node = stnode.node
@@ -92,7 +130,7 @@ class NodeFixer(NodeVisitor[None]):
                 elif value.node is not None:
                     value.node.accept(self)
                 else:
-                    assert False, f'Unexpected empty node {key!r}: {value}'
+                    assert False, f"Unexpected empty node {key!r}: {value}"
 
     def visit_func_def(self, func: FuncDef) -> None:
         if self.current_info is not None:
@@ -132,6 +170,12 @@ class NodeFixer(NodeVisitor[None]):
             value.accept(self.type_fixer)
         tv.upper_bound.accept(self.type_fixer)
 
+    def visit_paramspec_expr(self, p: ParamSpecExpr) -> None:
+        p.upper_bound.accept(self.type_fixer)
+
+    def visit_type_var_tuple_expr(self, tv: TypeVarTupleExpr) -> None:
+        tv.upper_bound.accept(self.type_fixer)
+
     def visit_var(self, v: Var) -> None:
         if self.current_info is not None:
             v.info = self.current_info
@@ -140,10 +184,12 @@ class NodeFixer(NodeVisitor[None]):
 
     def visit_type_alias(self, a: TypeAlias) -> None:
         a.target.accept(self.type_fixer)
+        for v in a.alias_tvars:
+            v.accept(self.type_fixer)
 
 
 class TypeFixer(TypeVisitor[None]):
-    def __init__(self, modules: Dict[str, MypyFile], allow_missing: bool) -> None:
+    def __init__(self, modules: dict[str, MypyFile], allow_missing: bool) -> None:
         self.modules = modules
         self.allow_missing = allow_missing
 
@@ -153,8 +199,9 @@ class TypeFixer(TypeVisitor[None]):
         if type_ref is None:
             return  # We've already been here.
         inst.type_ref = None
-        inst.type = lookup_fully_qualified_typeinfo(self.modules, type_ref,
-                                                    allow_missing=self.allow_missing)
+        inst.type = lookup_fully_qualified_typeinfo(
+            self.modules, type_ref, allow_missing=self.allow_missing
+        )
         # TODO: Is this needed or redundant?
         # Also fix up the bases, just in case.
         for base in inst.type.bases:
@@ -170,8 +217,9 @@ class TypeFixer(TypeVisitor[None]):
         if type_ref is None:
             return  # We've already been here.
         t.type_ref = None
-        t.alias = lookup_fully_qualified_alias(self.modules, type_ref,
-                                               allow_missing=self.allow_missing)
+        t.alias = lookup_fully_qualified_alias(
+            self.modules, type_ref, allow_missing=self.allow_missing
+        )
         for a in t.args:
             a.accept(self)
 
@@ -228,11 +276,17 @@ class TypeFixer(TypeVisitor[None]):
                 it.accept(self)
         if tdt.fallback is not None:
             if tdt.fallback.type_ref is not None:
-                if lookup_fully_qualified(tdt.fallback.type_ref, self.modules,
-                                          raise_on_missing=not self.allow_missing) is None:
+                if (
+                    lookup_fully_qualified(
+                        tdt.fallback.type_ref,
+                        self.modules,
+                        raise_on_missing=not self.allow_missing,
+                    )
+                    is None
+                ):
                     # We reject fake TypeInfos for TypedDict fallbacks because
                     # the latter are used in type checking and must be valid.
-                    tdt.fallback.type_ref = 'typing._TypedDict'
+                    tdt.fallback.type_ref = "typing._TypedDict"
             tdt.fallback.accept(self)
 
     def visit_literal_type(self, lt: LiteralType) -> None:
@@ -277,52 +331,68 @@ class TypeFixer(TypeVisitor[None]):
         t.item.accept(self)
 
 
-def lookup_fully_qualified_typeinfo(modules: Dict[str, MypyFile], name: str, *,
-                                    allow_missing: bool) -> TypeInfo:
+def lookup_fully_qualified_typeinfo(
+    modules: dict[str, MypyFile], name: str, *, allow_missing: bool
+) -> TypeInfo:
     stnode = lookup_fully_qualified(name, modules, raise_on_missing=not allow_missing)
     node = stnode.node if stnode else None
     if isinstance(node, TypeInfo):
         return node
     else:
         # Looks like a missing TypeInfo during an initial daemon load, put something there
-        assert allow_missing, "Should never get here in normal mode," \
-                              " got {}:{} instead of TypeInfo".format(type(node).__name__,
-                                                                      node.fullname if node
-                                                                      else '')
+        assert (
+            allow_missing
+        ), "Should never get here in normal mode, got {}:{} instead of TypeInfo".format(
+            type(node).__name__, node.fullname if node else ""
+        )
         return missing_info(modules)
 
 
-def lookup_fully_qualified_alias(modules: Dict[str, MypyFile], name: str, *,
-                                 allow_missing: bool) -> TypeAlias:
+def lookup_fully_qualified_alias(
+    modules: dict[str, MypyFile], name: str, *, allow_missing: bool
+) -> TypeAlias:
     stnode = lookup_fully_qualified(name, modules, raise_on_missing=not allow_missing)
     node = stnode.node if stnode else None
     if isinstance(node, TypeAlias):
         return node
+    elif isinstance(node, TypeInfo):
+        if node.special_alias:
+            # Already fixed up.
+            return node.special_alias
+        if node.tuple_type:
+            alias = TypeAlias.from_tuple_type(node)
+        elif node.typeddict_type:
+            alias = TypeAlias.from_typeddict_type(node)
+        else:
+            assert allow_missing
+            return missing_alias()
+        node.special_alias = alias
+        return alias
     else:
         # Looks like a missing TypeAlias during an initial daemon load, put something there
-        assert allow_missing, "Should never get here in normal mode," \
-                              " got {}:{} instead of TypeAlias".format(type(node).__name__,
-                                                                       node.fullname if node
-                                                                       else '')
+        assert (
+            allow_missing
+        ), "Should never get here in normal mode, got {}:{} instead of TypeAlias".format(
+            type(node).__name__, node.fullname if node else ""
+        )
         return missing_alias()
 
 
 _SUGGESTION: Final = "<missing {}: *should* have gone away during fine-grained update>"
 
 
-def missing_info(modules: Dict[str, MypyFile]) -> TypeInfo:
-    suggestion = _SUGGESTION.format('info')
+def missing_info(modules: dict[str, MypyFile]) -> TypeInfo:
+    suggestion = _SUGGESTION.format("info")
     dummy_def = ClassDef(suggestion, Block([]))
     dummy_def.fullname = suggestion
 
     info = TypeInfo(SymbolTable(), dummy_def, "<missing>")
-    obj_type = lookup_fully_qualified_typeinfo(modules, 'builtins.object', allow_missing=False)
+    obj_type = lookup_fully_qualified_typeinfo(modules, "builtins.object", allow_missing=False)
     info.bases = [Instance(obj_type, [])]
     info.mro = [info, obj_type]
     return info
 
 
 def missing_alias() -> TypeAlias:
-    suggestion = _SUGGESTION.format('alias')
-    return TypeAlias(AnyType(TypeOfAny.special_form), suggestion,
-                     line=-1, column=-1)
+    suggestion = _SUGGESTION.format("alias")
+    return TypeAlias(AnyType(TypeOfAny.special_form), suggestion, line=-1, column=-1)
