@@ -1,25 +1,42 @@
-from typing import List, Optional, Union
+from __future__ import annotations
 
+from mypy.fixup import TypeFixer
 from mypy.nodes import (
-    ARG_POS, MDEF, Argument, Block, CallExpr, ClassDef, Expression, SYMBOL_FUNCBASE_TYPES,
-    FuncDef, PassStmt, RefExpr, SymbolTableNode, Var, JsonDict,
+    ARG_POS,
+    MDEF,
+    SYMBOL_FUNCBASE_TYPES,
+    Argument,
+    Block,
+    CallExpr,
+    ClassDef,
+    Decorator,
+    Expression,
+    FuncDef,
+    JsonDict,
+    PassStmt,
+    RefExpr,
+    SymbolTableNode,
+    Var,
 )
 from mypy.plugin import CheckerPluginInterface, ClassDefContext, SemanticAnalyzerPluginInterface
-from mypy.semanal import set_callable_name
+from mypy.semanal_shared import ALLOW_INCOMPATIBLE_OVERRIDE, set_callable_name
+from mypy.typeops import (  # noqa: F401  # Part of public API
+    try_getting_str_literals as try_getting_str_literals,
+)
 from mypy.types import (
-    CallableType, Overloaded, Type, TypeVarType, deserialize_type, get_proper_type,
+    CallableType,
+    Overloaded,
+    Type,
+    TypeType,
+    TypeVarType,
+    deserialize_type,
+    get_proper_type,
 )
 from mypy.typevars import fill_typevars
 from mypy.util import get_unique_redefinition_name
-from mypy.typeops import try_getting_str_literals  # noqa: F401  # Part of public API
-from mypy.fixup import TypeFixer
 
 
-def _get_decorator_bool_argument(
-        ctx: ClassDefContext,
-        name: str,
-        default: bool,
-) -> bool:
+def _get_decorator_bool_argument(ctx: ClassDefContext, name: str, default: bool) -> bool:
     """Return the bool argument for the decorator.
 
     This handles both @decorator(...) and @decorator.
@@ -30,8 +47,7 @@ def _get_decorator_bool_argument(
         return default
 
 
-def _get_bool_argument(ctx: ClassDefContext, expr: CallExpr,
-                       name: str, default: bool) -> bool:
+def _get_bool_argument(ctx: ClassDefContext, expr: CallExpr, name: str, default: bool) -> bool:
     """Return the boolean value for an argument to a call or the
     default if it's not found.
     """
@@ -39,13 +55,13 @@ def _get_bool_argument(ctx: ClassDefContext, expr: CallExpr,
     if attr_value:
         ret = ctx.api.parse_bool(attr_value)
         if ret is None:
-            ctx.api.fail('"{}" argument must be True or False.'.format(name), expr)
+            ctx.api.fail(f'"{name}" argument must be True or False.', expr)
             return default
         return ret
     return default
 
 
-def _get_argument(call: CallExpr, name: str) -> Optional[Expression]:
+def _get_argument(call: CallExpr, name: str) -> Expression | None:
     """Return the expression for the specific argument."""
     # To do this we use the CallableType of the callee to find the FormalArgument,
     # then walk the actual CallExpr looking for the appropriate argument.
@@ -57,8 +73,7 @@ def _get_argument(call: CallExpr, name: str) -> Optional[Expression]:
 
     callee_type = None
     callee_node = call.callee.node
-    if (isinstance(callee_node, (Var, SYMBOL_FUNCBASE_TYPES))
-            and callee_node.type):
+    if isinstance(callee_node, (Var, SYMBOL_FUNCBASE_TYPES)) and callee_node.type:
         callee_node_type = get_proper_type(callee_node.type)
         if isinstance(callee_node_type, Overloaded):
             # We take the last overload.
@@ -83,35 +98,49 @@ def _get_argument(call: CallExpr, name: str) -> Optional[Expression]:
 
 
 def add_method(
-        ctx: ClassDefContext,
-        name: str,
-        args: List[Argument],
-        return_type: Type,
-        self_type: Optional[Type] = None,
-        tvar_def: Optional[TypeVarType] = None,
+    ctx: ClassDefContext,
+    name: str,
+    args: list[Argument],
+    return_type: Type,
+    self_type: Type | None = None,
+    tvar_def: TypeVarType | None = None,
+    is_classmethod: bool = False,
+    is_staticmethod: bool = False,
 ) -> None:
     """
     Adds a new method to a class.
     Deprecated, use add_method_to_class() instead.
     """
-    add_method_to_class(ctx.api, ctx.cls,
-                        name=name,
-                        args=args,
-                        return_type=return_type,
-                        self_type=self_type,
-                        tvar_def=tvar_def)
+    add_method_to_class(
+        ctx.api,
+        ctx.cls,
+        name=name,
+        args=args,
+        return_type=return_type,
+        self_type=self_type,
+        tvar_def=tvar_def,
+        is_classmethod=is_classmethod,
+        is_staticmethod=is_staticmethod,
+    )
 
 
 def add_method_to_class(
-        api: Union[SemanticAnalyzerPluginInterface, CheckerPluginInterface],
-        cls: ClassDef,
-        name: str,
-        args: List[Argument],
-        return_type: Type,
-        self_type: Optional[Type] = None,
-        tvar_def: Optional[TypeVarType] = None,
+    api: SemanticAnalyzerPluginInterface | CheckerPluginInterface,
+    cls: ClassDef,
+    name: str,
+    args: list[Argument],
+    return_type: Type,
+    self_type: Type | None = None,
+    tvar_def: TypeVarType | None = None,
+    is_classmethod: bool = False,
+    is_staticmethod: bool = False,
 ) -> None:
     """Adds a new method to a class definition."""
+
+    assert not (
+        is_classmethod is True and is_staticmethod is True
+    ), "Can't add a new method that's both staticmethod and classmethod."
+
     info = cls.info
 
     # First remove any previously generated methods with the same name
@@ -121,16 +150,24 @@ def add_method_to_class(
         if sym.plugin_generated and isinstance(sym.node, FuncDef):
             cls.defs.body.remove(sym.node)
 
-    self_type = self_type or fill_typevars(info)
     if isinstance(api, SemanticAnalyzerPluginInterface):
-        function_type = api.named_type('builtins.function')
+        function_type = api.named_type("builtins.function")
     else:
-        function_type = api.named_generic_type('builtins.function', [])
+        function_type = api.named_generic_type("builtins.function", [])
+
+    if is_classmethod:
+        self_type = self_type or TypeType(fill_typevars(info))
+        first = [Argument(Var("_cls"), self_type, None, ARG_POS, True)]
+    elif is_staticmethod:
+        first = []
+    else:
+        self_type = self_type or fill_typevars(info)
+        first = [Argument(Var("self"), self_type, None, ARG_POS)]
+    args = first + args
 
-    args = [Argument(Var('self'), self_type, None, ARG_POS)] + args
     arg_types, arg_names, arg_kinds = [], [], []
     for arg in args:
-        assert arg.type_annotation, 'All arguments must be fully typed.'
+        assert arg.type_annotation, "All arguments must be fully typed."
         arg_types.append(arg.type_annotation)
         arg_names.append(arg.variable.name)
         arg_kinds.append(arg.kind)
@@ -142,7 +179,9 @@ def add_method_to_class(
     func = FuncDef(name, args, Block([PassStmt()]))
     func.info = info
     func.type = set_callable_name(signature, func)
-    func._fullname = info.fullname + '.' + name
+    func.is_class = is_classmethod
+    func.is_static = is_staticmethod
+    func._fullname = info.fullname + "." + name
     func.line = info.line
 
     # NOTE: we would like the plugin generated node to dominate, but we still
@@ -152,13 +191,68 @@ def add_method_to_class(
         r_name = get_unique_redefinition_name(name, info.names)
         info.names[r_name] = info.names[name]
 
-    info.names[name] = SymbolTableNode(MDEF, func, plugin_generated=True)
+    # Add decorator for is_staticmethod. It's unnecessary for is_classmethod.
+    if is_staticmethod:
+        func.is_decorated = True
+        v = Var(name, func.type)
+        v.info = info
+        v._fullname = func._fullname
+        v.is_staticmethod = True
+        dec = Decorator(func, [], v)
+        dec.line = info.line
+        sym = SymbolTableNode(MDEF, dec)
+    else:
+        sym = SymbolTableNode(MDEF, func)
+    sym.plugin_generated = True
+    info.names[name] = sym
+
     info.defn.defs.body.append(func)
 
 
-def deserialize_and_fixup_type(
-    data: Union[str, JsonDict], api: SemanticAnalyzerPluginInterface
-) -> Type:
+def add_attribute_to_class(
+    api: SemanticAnalyzerPluginInterface,
+    cls: ClassDef,
+    name: str,
+    typ: Type,
+    final: bool = False,
+    no_serialize: bool = False,
+    override_allow_incompatible: bool = False,
+    fullname: str | None = None,
+    is_classvar: bool = False,
+) -> None:
+    """
+    Adds a new attribute to a class definition.
+    This currently only generates the symbol table entry and no corresponding AssignmentStatement
+    """
+    info = cls.info
+
+    # NOTE: we would like the plugin generated node to dominate, but we still
+    # need to keep any existing definitions so they get semantically analyzed.
+    if name in info.names:
+        # Get a nice unique name instead.
+        r_name = get_unique_redefinition_name(name, info.names)
+        info.names[r_name] = info.names[name]
+
+    node = Var(name, typ)
+    node.info = info
+    node.is_final = final
+    node.is_classvar = is_classvar
+    if name in ALLOW_INCOMPATIBLE_OVERRIDE:
+        node.allow_incompatible_override = True
+    else:
+        node.allow_incompatible_override = override_allow_incompatible
+
+    if fullname:
+        node._fullname = fullname
+    else:
+        node._fullname = info.fullname + "." + name
+
+    info.names[name] = SymbolTableNode(
+        MDEF, node, plugin_generated=True, no_serialize=no_serialize
+    )
+
+
+def deserialize_and_fixup_type(data: str | JsonDict, api: SemanticAnalyzerPluginInterface) -> Type:
     typ = deserialize_type(data)
     typ.accept(TypeFixer(api.modules, allow_missing=False))
     return typ
