@@ -1,75 +1,89 @@
+from __future__ import annotations
+
 from functools import partial
-from typing import Callable, Optional, List
+from typing import Callable
 
 from mypy import message_registry
-from mypy.nodes import StrExpr, IntExpr, DictExpr, UnaryExpr
+from mypy.nodes import DictExpr, IntExpr, StrExpr, UnaryExpr
 from mypy.plugin import (
-    Plugin, FunctionContext, MethodContext, MethodSigContext, AttributeContext, ClassDefContext
+    AttributeContext,
+    ClassDefContext,
+    FunctionContext,
+    MethodContext,
+    MethodSigContext,
+    Plugin,
 )
 from mypy.plugins.common import try_getting_str_literals
-from mypy.types import (
-    FunctionLike, Type, Instance, AnyType, TypeOfAny, CallableType, NoneType, TypedDictType,
-    TypeVarType, TPDICT_FB_NAMES, get_proper_type, LiteralType, TupleType
-)
 from mypy.subtypes import is_subtype
-from mypy.typeops import make_simplified_union
-from mypy.checkexpr import is_literal_type_like
+from mypy.typeops import is_literal_type_like, make_simplified_union
+from mypy.types import (
+    TPDICT_FB_NAMES,
+    AnyType,
+    CallableType,
+    FunctionLike,
+    Instance,
+    LiteralType,
+    NoneType,
+    TupleType,
+    Type,
+    TypedDictType,
+    TypeOfAny,
+    TypeVarType,
+    get_proper_type,
+)
 
 
 class DefaultPlugin(Plugin):
     """Type checker plugin that is enabled by default."""
 
-    def get_function_hook(self, fullname: str
-                          ) -> Optional[Callable[[FunctionContext], Type]]:
+    def get_function_hook(self, fullname: str) -> Callable[[FunctionContext], Type] | None:
         from mypy.plugins import ctypes, singledispatch
 
-        if fullname in ('contextlib.contextmanager', 'contextlib.asynccontextmanager'):
-            return contextmanager_callback
-        elif fullname == 'ctypes.Array':
+        if fullname == "ctypes.Array":
             return ctypes.array_constructor_callback
-        elif fullname == 'functools.singledispatch':
+        elif fullname == "functools.singledispatch":
             return singledispatch.create_singledispatch_function_callback
         return None
 
-    def get_method_signature_hook(self, fullname: str
-                                  ) -> Optional[Callable[[MethodSigContext], FunctionLike]]:
+    def get_method_signature_hook(
+        self, fullname: str
+    ) -> Callable[[MethodSigContext], FunctionLike] | None:
         from mypy.plugins import ctypes, singledispatch
 
-        if fullname == 'typing.Mapping.get':
+        if fullname == "typing.Mapping.get":
             return typed_dict_get_signature_callback
-        elif fullname in set(n + '.setdefault' for n in TPDICT_FB_NAMES):
+        elif fullname in {n + ".setdefault" for n in TPDICT_FB_NAMES}:
             return typed_dict_setdefault_signature_callback
-        elif fullname in set(n + '.pop' for n in TPDICT_FB_NAMES):
+        elif fullname in {n + ".pop" for n in TPDICT_FB_NAMES}:
             return typed_dict_pop_signature_callback
-        elif fullname in set(n + '.update' for n in TPDICT_FB_NAMES):
+        elif fullname in {n + ".update" for n in TPDICT_FB_NAMES}:
             return typed_dict_update_signature_callback
-        elif fullname == 'ctypes.Array.__setitem__':
+        elif fullname == "ctypes.Array.__setitem__":
             return ctypes.array_setitem_callback
         elif fullname == singledispatch.SINGLEDISPATCH_CALLABLE_CALL_METHOD:
             return singledispatch.call_singledispatch_function_callback
         return None
 
-    def get_method_hook(self, fullname: str
-                        ) -> Optional[Callable[[MethodContext], Type]]:
+    def get_method_hook(self, fullname: str) -> Callable[[MethodContext], Type] | None:
         from mypy.plugins import ctypes, singledispatch
 
-        if fullname == 'typing.Mapping.get':
+        if fullname == "typing.Mapping.get":
             return typed_dict_get_callback
-        elif fullname == 'builtins.int.__pow__':
+        elif fullname == "builtins.int.__pow__":
             return int_pow_callback
-        elif fullname == 'builtins.int.__neg__':
+        elif fullname == "builtins.int.__neg__":
             return int_neg_callback
-        elif fullname in ('builtins.tuple.__mul__', 'builtins.tuple.__rmul__'):
+        elif fullname in ("builtins.tuple.__mul__", "builtins.tuple.__rmul__"):
             return tuple_mul_callback
-        elif fullname in set(n + '.setdefault' for n in TPDICT_FB_NAMES):
+        elif fullname in {n + ".setdefault" for n in TPDICT_FB_NAMES}:
             return typed_dict_setdefault_callback
-        elif fullname in set(n + '.pop' for n in TPDICT_FB_NAMES):
+        elif fullname in {n + ".pop" for n in TPDICT_FB_NAMES}:
             return typed_dict_pop_callback
-        elif fullname in set(n + '.__delitem__' for n in TPDICT_FB_NAMES):
+        elif fullname in {n + ".__delitem__" for n in TPDICT_FB_NAMES}:
             return typed_dict_delitem_callback
-        elif fullname == 'ctypes.Array.__getitem__':
+        elif fullname == "ctypes.Array.__getitem__":
             return ctypes.array_getitem_callback
-        elif fullname == 'ctypes.Array.__iter__':
+        elif fullname == "ctypes.Array.__iter__":
             return ctypes.array_iter_callback
         elif fullname == singledispatch.SINGLEDISPATCH_REGISTER_METHOD:
             return singledispatch.singledispatch_register_callback
@@ -77,14 +91,12 @@ class DefaultPlugin(Plugin):
             return singledispatch.call_singledispatch_function_after_register_argument
         return None
 
-    def get_attribute_hook(self, fullname: str
-                           ) -> Optional[Callable[[AttributeContext], Type]]:
-        from mypy.plugins import ctypes
-        from mypy.plugins import enums
+    def get_attribute_hook(self, fullname: str) -> Callable[[AttributeContext], Type] | None:
+        from mypy.plugins import ctypes, enums
 
-        if fullname == 'ctypes.Array.value':
+        if fullname == "ctypes.Array.value":
             return ctypes.array_value_callback
-        elif fullname == 'ctypes.Array.raw':
+        elif fullname == "ctypes.Array.raw":
             return ctypes.array_raw_callback
         elif fullname in enums.ENUM_NAME_ACCESS:
             return enums.enum_name_callback
@@ -92,55 +104,46 @@ class DefaultPlugin(Plugin):
             return enums.enum_value_callback
         return None
 
-    def get_class_decorator_hook(self, fullname: str
-                                 ) -> Optional[Callable[[ClassDefContext], None]]:
-        from mypy.plugins import attrs
-        from mypy.plugins import dataclasses
-        from mypy.plugins import functools
+    def get_class_decorator_hook(self, fullname: str) -> Callable[[ClassDefContext], None] | None:
+        from mypy.plugins import attrs, dataclasses
 
-        if fullname in attrs.attr_class_makers:
-            return attrs.attr_class_maker_callback
-        elif fullname in attrs.attr_dataclass_makers:
-            return partial(
-                attrs.attr_class_maker_callback,
-                auto_attribs_default=True,
-            )
-        elif fullname in attrs.attr_frozen_makers:
-            return partial(
-                attrs.attr_class_maker_callback,
-                auto_attribs_default=None,
-                frozen_default=True,
-            )
-        elif fullname in attrs.attr_define_makers:
-            return partial(
-                attrs.attr_class_maker_callback,
-                auto_attribs_default=None,
-            )
-        elif fullname in dataclasses.dataclass_makers:
-            return dataclasses.dataclass_class_maker_callback
-        elif fullname in functools.functools_total_ordering_makers:
-            return functools.functools_total_ordering_maker_callback
+        # These dataclass and attrs hooks run in the main semantic analysis pass
+        # and only tag known dataclasses/attrs classes, so that the second
+        # hooks (in get_class_decorator_hook_2) can detect dataclasses/attrs classes
+        # in the MRO.
+        if fullname in dataclasses.dataclass_makers:
+            return dataclasses.dataclass_tag_callback
+        if (
+            fullname in attrs.attr_class_makers
+            or fullname in attrs.attr_dataclass_makers
+            or fullname in attrs.attr_frozen_makers
+            or fullname in attrs.attr_define_makers
+        ):
+            return attrs.attr_tag_callback
 
         return None
 
+    def get_class_decorator_hook_2(
+        self, fullname: str
+    ) -> Callable[[ClassDefContext], bool] | None:
+        from mypy.plugins import attrs, dataclasses, functools
 
-def contextmanager_callback(ctx: FunctionContext) -> Type:
-    """Infer a better return type for 'contextlib.contextmanager'."""
-    # Be defensive, just in case.
-    if ctx.arg_types and len(ctx.arg_types[0]) == 1:
-        arg_type = get_proper_type(ctx.arg_types[0][0])
-        default_return = get_proper_type(ctx.default_return_type)
-        if (isinstance(arg_type, CallableType)
-                and isinstance(default_return, CallableType)):
-            # The stub signature doesn't preserve information about arguments so
-            # add them back here.
-            return default_return.copy_modified(
-                arg_types=arg_type.arg_types,
-                arg_kinds=arg_type.arg_kinds,
-                arg_names=arg_type.arg_names,
-                variables=arg_type.variables,
-                is_ellipsis_args=arg_type.is_ellipsis_args)
-    return ctx.default_return_type
+        if fullname in dataclasses.dataclass_makers:
+            return dataclasses.dataclass_class_maker_callback
+        elif fullname in functools.functools_total_ordering_makers:
+            return functools.functools_total_ordering_maker_callback
+        elif fullname in attrs.attr_class_makers:
+            return attrs.attr_class_maker_callback
+        elif fullname in attrs.attr_dataclass_makers:
+            return partial(attrs.attr_class_maker_callback, auto_attribs_default=True)
+        elif fullname in attrs.attr_frozen_makers:
+            return partial(
+                attrs.attr_class_maker_callback, auto_attribs_default=None, frozen_default=True
+            )
+        elif fullname in attrs.attr_define_makers:
+            return partial(attrs.attr_class_maker_callback, auto_attribs_default=None)
+
+        return None
 
 
 def typed_dict_get_signature_callback(ctx: MethodSigContext) -> CallableType:
@@ -150,21 +153,25 @@ def typed_dict_get_signature_callback(ctx: MethodSigContext) -> CallableType:
     depends on a TypedDict value type.
     """
     signature = ctx.default_signature
-    if (isinstance(ctx.type, TypedDictType)
-            and len(ctx.args) == 2
-            and len(ctx.args[0]) == 1
-            and isinstance(ctx.args[0][0], StrExpr)
-            and len(signature.arg_types) == 2
-            and len(signature.variables) == 1
-            and len(ctx.args[1]) == 1):
+    if (
+        isinstance(ctx.type, TypedDictType)
+        and len(ctx.args) == 2
+        and len(ctx.args[0]) == 1
+        and isinstance(ctx.args[0][0], StrExpr)
+        and len(signature.arg_types) == 2
+        and len(signature.variables) == 1
+        and len(ctx.args[1]) == 1
+    ):
         key = ctx.args[0][0].value
         value_type = get_proper_type(ctx.type.items.get(key))
         ret_type = signature.ret_type
         if value_type:
             default_arg = ctx.args[1][0]
-            if (isinstance(value_type, TypedDictType)
-                    and isinstance(default_arg, DictExpr)
-                    and len(default_arg.items) == 0):
+            if (
+                isinstance(value_type, TypedDictType)
+                and isinstance(default_arg, DictExpr)
+                and len(default_arg.items) == 0
+            ):
                 # Caller has empty dict {} as default for typed dict.
                 value_type = value_type.copy_modified(required_keys=set())
             # Tweak the signature to include the value type as context. It's
@@ -173,22 +180,24 @@ def typed_dict_get_signature_callback(ctx: MethodSigContext) -> CallableType:
             tv = signature.variables[0]
             assert isinstance(tv, TypeVarType)
             return signature.copy_modified(
-                arg_types=[signature.arg_types[0],
-                           make_simplified_union([value_type, tv])],
-                ret_type=ret_type)
+                arg_types=[signature.arg_types[0], make_simplified_union([value_type, tv])],
+                ret_type=ret_type,
+            )
     return signature
 
 
 def typed_dict_get_callback(ctx: MethodContext) -> Type:
     """Infer a precise return type for TypedDict.get with literal first argument."""
-    if (isinstance(ctx.type, TypedDictType)
-            and len(ctx.arg_types) >= 1
-            and len(ctx.arg_types[0]) == 1):
+    if (
+        isinstance(ctx.type, TypedDictType)
+        and len(ctx.arg_types) >= 1
+        and len(ctx.arg_types[0]) == 1
+    ):
         keys = try_getting_str_literals(ctx.args[0][0], ctx.arg_types[0][0])
         if keys is None:
             return ctx.default_return_type
 
-        output_types: List[Type] = []
+        output_types: list[Type] = []
         for key in keys:
             value_type = get_proper_type(ctx.type.items.get(key))
             if value_type is None:
@@ -196,11 +205,13 @@ def typed_dict_get_callback(ctx: MethodContext) -> Type:
 
             if len(ctx.arg_types) == 1:
                 output_types.append(value_type)
-            elif (len(ctx.arg_types) == 2 and len(ctx.arg_types[1]) == 1
-                  and len(ctx.args[1]) == 1):
+            elif len(ctx.arg_types) == 2 and len(ctx.arg_types[1]) == 1 and len(ctx.args[1]) == 1:
                 default_arg = ctx.args[1][0]
-                if (isinstance(default_arg, DictExpr) and len(default_arg.items) == 0
-                        and isinstance(value_type, TypedDictType)):
+                if (
+                    isinstance(default_arg, DictExpr)
+                    and len(default_arg.items) == 0
+                    and isinstance(value_type, TypedDictType)
+                ):
                     # Special case '{}' as the default for a typed dict type.
                     output_types.append(value_type.copy_modified(required_keys=set()))
                 else:
@@ -221,14 +232,16 @@ def typed_dict_pop_signature_callback(ctx: MethodSigContext) -> CallableType:
     depends on a TypedDict value type.
     """
     signature = ctx.default_signature
-    str_type = ctx.api.named_generic_type('builtins.str', [])
-    if (isinstance(ctx.type, TypedDictType)
-            and len(ctx.args) == 2
-            and len(ctx.args[0]) == 1
-            and isinstance(ctx.args[0][0], StrExpr)
-            and len(signature.arg_types) == 2
-            and len(signature.variables) == 1
-            and len(ctx.args[1]) == 1):
+    str_type = ctx.api.named_generic_type("builtins.str", [])
+    if (
+        isinstance(ctx.type, TypedDictType)
+        and len(ctx.args) == 2
+        and len(ctx.args[0]) == 1
+        and isinstance(ctx.args[0][0], StrExpr)
+        and len(signature.arg_types) == 2
+        and len(signature.variables) == 1
+        and len(ctx.args[1]) == 1
+    ):
         key = ctx.args[0][0].value
         value_type = ctx.type.items.get(key)
         if value_type:
@@ -238,17 +251,17 @@ def typed_dict_pop_signature_callback(ctx: MethodSigContext) -> CallableType:
             tv = signature.variables[0]
             assert isinstance(tv, TypeVarType)
             typ = make_simplified_union([value_type, tv])
-            return signature.copy_modified(
-                arg_types=[str_type, typ],
-                ret_type=typ)
+            return signature.copy_modified(arg_types=[str_type, typ], ret_type=typ)
     return signature.copy_modified(arg_types=[str_type, signature.arg_types[1]])
 
 
 def typed_dict_pop_callback(ctx: MethodContext) -> Type:
     """Type check and infer a precise return type for TypedDict.pop."""
-    if (isinstance(ctx.type, TypedDictType)
-            and len(ctx.arg_types) >= 1
-            and len(ctx.arg_types[0]) == 1):
+    if (
+        isinstance(ctx.type, TypedDictType)
+        and len(ctx.arg_types) >= 1
+        and len(ctx.arg_types[0]) == 1
+    ):
         keys = try_getting_str_literals(ctx.args[0][0], ctx.arg_types[0][0])
         if keys is None:
             ctx.api.fail(message_registry.TYPEDDICT_KEY_MUST_BE_STRING_LITERAL, ctx.context)
@@ -268,8 +281,7 @@ def typed_dict_pop_callback(ctx: MethodContext) -> Type:
 
         if len(ctx.args[1]) == 0:
             return make_simplified_union(value_types)
-        elif (len(ctx.arg_types) == 2 and len(ctx.arg_types[1]) == 1
-              and len(ctx.args[1]) == 1):
+        elif len(ctx.arg_types) == 2 and len(ctx.arg_types[1]) == 1 and len(ctx.args[1]) == 1:
             return make_simplified_union([*value_types, ctx.arg_types[1][0]])
     return ctx.default_return_type
 
@@ -281,13 +293,15 @@ def typed_dict_setdefault_signature_callback(ctx: MethodSigContext) -> CallableT
     depends on a TypedDict value type.
     """
     signature = ctx.default_signature
-    str_type = ctx.api.named_generic_type('builtins.str', [])
-    if (isinstance(ctx.type, TypedDictType)
-            and len(ctx.args) == 2
-            and len(ctx.args[0]) == 1
-            and isinstance(ctx.args[0][0], StrExpr)
-            and len(signature.arg_types) == 2
-            and len(ctx.args[1]) == 1):
+    str_type = ctx.api.named_generic_type("builtins.str", [])
+    if (
+        isinstance(ctx.type, TypedDictType)
+        and len(ctx.args) == 2
+        and len(ctx.args[0]) == 1
+        and isinstance(ctx.args[0][0], StrExpr)
+        and len(signature.arg_types) == 2
+        and len(ctx.args[1]) == 1
+    ):
         key = ctx.args[0][0].value
         value_type = ctx.type.items.get(key)
         if value_type:
@@ -297,10 +311,12 @@ def typed_dict_setdefault_signature_callback(ctx: MethodSigContext) -> CallableT
 
 def typed_dict_setdefault_callback(ctx: MethodContext) -> Type:
     """Type check TypedDict.setdefault and infer a precise return type."""
-    if (isinstance(ctx.type, TypedDictType)
-            and len(ctx.arg_types) == 2
-            and len(ctx.arg_types[0]) == 1
-            and len(ctx.arg_types[1]) == 1):
+    if (
+        isinstance(ctx.type, TypedDictType)
+        and len(ctx.arg_types) == 2
+        and len(ctx.arg_types[0]) == 1
+        and len(ctx.arg_types[1]) == 1
+    ):
         keys = try_getting_str_literals(ctx.args[0][0], ctx.arg_types[0][0])
         if keys is None:
             ctx.api.fail(message_registry.TYPEDDICT_KEY_MUST_BE_STRING_LITERAL, ctx.context)
@@ -322,7 +338,8 @@ def typed_dict_setdefault_callback(ctx: MethodContext) -> Type:
             # default can be assigned to all key-value pairs we're updating.
             if not is_subtype(default_type, value_type):
                 ctx.api.msg.typeddict_setdefault_arguments_inconsistent(
-                    default_type, value_type, ctx.context)
+                    default_type, value_type, ctx.context
+                )
                 return AnyType(TypeOfAny.from_error)
 
             value_types.append(value_type)
@@ -333,9 +350,11 @@ def typed_dict_setdefault_callback(ctx: MethodContext) -> Type:
 
 def typed_dict_delitem_callback(ctx: MethodContext) -> Type:
     """Type check TypedDict.__delitem__."""
-    if (isinstance(ctx.type, TypedDictType)
-            and len(ctx.arg_types) == 1
-            and len(ctx.arg_types[0]) == 1):
+    if (
+        isinstance(ctx.type, TypedDictType)
+        and len(ctx.arg_types) == 1
+        and len(ctx.arg_types[0]) == 1
+    ):
         keys = try_getting_str_literals(ctx.args[0][0], ctx.arg_types[0][0])
         if keys is None:
             ctx.api.fail(message_registry.TYPEDDICT_KEY_MUST_BE_STRING_LITERAL, ctx.context)
@@ -352,8 +371,7 @@ def typed_dict_delitem_callback(ctx: MethodContext) -> Type:
 def typed_dict_update_signature_callback(ctx: MethodSigContext) -> CallableType:
     """Try to infer a better signature type for TypedDict.update."""
     signature = ctx.default_signature
-    if (isinstance(ctx.type, TypedDictType)
-            and len(signature.arg_types) == 1):
+    if isinstance(ctx.type, TypedDictType) and len(signature.arg_types) == 1:
         arg_type = get_proper_type(signature.arg_types[0])
         assert isinstance(arg_type, TypedDictType)
         arg_type = arg_type.as_anonymous()
@@ -366,20 +384,19 @@ def int_pow_callback(ctx: MethodContext) -> Type:
     """Infer a more precise return type for int.__pow__."""
     # int.__pow__ has an optional modulo argument,
     # so we expect 2 argument positions
-    if (len(ctx.arg_types) == 2
-            and len(ctx.arg_types[0]) == 1 and len(ctx.arg_types[1]) == 0):
+    if len(ctx.arg_types) == 2 and len(ctx.arg_types[0]) == 1 and len(ctx.arg_types[1]) == 0:
         arg = ctx.args[0][0]
         if isinstance(arg, IntExpr):
             exponent = arg.value
-        elif isinstance(arg, UnaryExpr) and arg.op == '-' and isinstance(arg.expr, IntExpr):
+        elif isinstance(arg, UnaryExpr) and arg.op == "-" and isinstance(arg.expr, IntExpr):
             exponent = -arg.expr.value
         else:
             # Right operand not an int literal or a negated literal -- give up.
             return ctx.default_return_type
         if exponent >= 0:
-            return ctx.api.named_generic_type('builtins.int', [])
+            return ctx.api.named_generic_type("builtins.int", [])
         else:
-            return ctx.api.named_generic_type('builtins.float', [])
+            return ctx.api.named_generic_type("builtins.float", [])
     return ctx.default_return_type
 
 
@@ -396,12 +413,11 @@ def int_neg_callback(ctx: MethodContext) -> Type:
             if is_literal_type_like(ctx.api.type_context[-1]):
                 return LiteralType(value=-value, fallback=fallback)
             else:
-                return ctx.type.copy_modified(last_known_value=LiteralType(
-                    value=-value,
-                    fallback=ctx.type,
-                    line=ctx.type.line,
-                    column=ctx.type.column,
-                ))
+                return ctx.type.copy_modified(
+                    last_known_value=LiteralType(
+                        value=-value, fallback=ctx.type, line=ctx.type.line, column=ctx.type.column
+                    )
+                )
     elif isinstance(ctx.type, LiteralType):
         value = ctx.type.value
         fallback = ctx.type.fallback
