@@ -1309,10 +1309,7 @@ class StubGenerator(mypy.traverser.TraverserVisitor):
 
     def is_private_member(self, fullname: str) -> bool:
         parts = fullname.split(".")
-        for part in parts:
-            if self.is_private_name(part):
-                return True
-        return False
+        return any(self.is_private_name(part) for part in parts)
 
     def get_str_type_of_node(
         self, rvalue: Expression, can_infer_optional: bool = False, can_be_any: bool = True
