@@ -202,7 +202,7 @@ class SemAnalTypeInfoSuite(DataSuite):
             for f in result.files.values():
                 for n in f.names.values():
                     if isinstance(n.node, TypeInfo):
-                        assert n.fullname is not None
+                        assert n.fullname
                         typeinfos[n.fullname] = n.node
 
             # The output is the symbol table converted into a string.
