@@ -1,17 +1,51 @@
 """Data-flow analyses."""
 
+from __future__ import annotations
+
 from abc import abstractmethod
+from typing import Dict, Generic, Iterable, Iterator, Set, Tuple, TypeVar
 
-from typing import Dict, Tuple, List, Set, TypeVar, Iterator, Generic, Optional, Iterable, Union
-
-from mypyc.ir.ops import (
-    Value, ControlOp,
-    BasicBlock, OpVisitor, Assign, Integer, LoadErrorValue, RegisterOp, Goto, Branch, Return, Call,
-    Box, Unbox, Cast, Op, Unreachable, TupleGet, TupleSet, GetAttr, SetAttr,
-    LoadStatic, InitStatic, MethodCall, RaiseStandardError, CallC, LoadGlobal,
-    Truncate, IntOp, LoadMem, GetElementPtr, LoadAddress, ComparisonOp, SetMem
-)
 from mypyc.ir.func_ir import all_values
+from mypyc.ir.ops import (
+    Assign,
+    AssignMulti,
+    BasicBlock,
+    Box,
+    Branch,
+    Call,
+    CallC,
+    Cast,
+    ComparisonOp,
+    ControlOp,
+    Extend,
+    GetAttr,
+    GetElementPtr,
+    Goto,
+    InitStatic,
+    Integer,
+    IntOp,
+    KeepAlive,
+    LoadAddress,
+    LoadErrorValue,
+    LoadGlobal,
+    LoadLiteral,
+    LoadMem,
+    LoadStatic,
+    MethodCall,
+    Op,
+    OpVisitor,
+    RaiseStandardError,
+    RegisterOp,
+    Return,
+    SetAttr,
+    SetMem,
+    Truncate,
+    TupleGet,
+    TupleSet,
+    Unbox,
+    Unreachable,
+    Value,
+)
 
 
 class CFG:
@@ -21,10 +55,12 @@ class CFG:
     non-empty set of exits.
     """
 
-    def __init__(self,
-                 succ: Dict[BasicBlock, List[BasicBlock]],
-                 pred: Dict[BasicBlock, List[BasicBlock]],
-                 exits: Set[BasicBlock]) -> None:
+    def __init__(
+        self,
+        succ: dict[BasicBlock, list[BasicBlock]],
+        pred: dict[BasicBlock, list[BasicBlock]],
+        exits: set[BasicBlock],
+    ) -> None:
         assert exits
         self.succ = succ
         self.pred = pred
@@ -32,13 +68,13 @@ class CFG:
 
     def __str__(self) -> str:
         lines = []
-        lines.append('exits: %s' % sorted(self.exits, key=lambda e: e.label))
-        lines.append('succ: %s' % self.succ)
-        lines.append('pred: %s' % self.pred)
-        return '\n'.join(lines)
+        lines.append("exits: %s" % sorted(self.exits, key=lambda e: int(e.label)))
+        lines.append("succ: %s" % self.succ)
+        lines.append("pred: %s" % self.pred)
+        return "\n".join(lines)
 
 
-def get_cfg(blocks: List[BasicBlock]) -> CFG:
+def get_cfg(blocks: list[BasicBlock]) -> CFG:
     """Calculate basic block control-flow graph.
 
     The result is a dictionary like this:
@@ -46,20 +82,16 @@ def get_cfg(blocks: List[BasicBlock]) -> CFG:
          basic block index -> (successors blocks, predecesssor blocks)
     """
     succ_map = {}
-    pred_map = {}  # type: Dict[BasicBlock, List[BasicBlock]]
+    pred_map: dict[BasicBlock, list[BasicBlock]] = {}
     exits = set()
     for block in blocks:
 
-        assert not any(isinstance(op, ControlOp) for op in block.ops[:-1]), (
-            "Control-flow ops must be at the end of blocks")
+        assert not any(
+            isinstance(op, ControlOp) for op in block.ops[:-1]
+        ), "Control-flow ops must be at the end of blocks"
 
-        last = block.ops[-1]
-        if isinstance(last, Branch):
-            succ = [last.true, last.false]
-        elif isinstance(last, Goto):
-            succ = [last.label]
-        else:
-            succ = []
+        succ = list(block.terminator.targets())
+        if not succ:
             exits.add(block)
 
         # Errors can occur anywhere inside a block, which means that
@@ -92,7 +124,7 @@ def get_real_target(label: BasicBlock) -> BasicBlock:
     return label
 
 
-def cleanup_cfg(blocks: List[BasicBlock]) -> None:
+def cleanup_cfg(blocks: list[BasicBlock]) -> None:
     """Cleanup the control flow graph.
 
     This eliminates obviously dead basic blocks and eliminates blocks that contain
@@ -104,12 +136,8 @@ def cleanup_cfg(blocks: List[BasicBlock]) -> None:
     while changed:
         # First collapse any jumps to basic block that only contain a goto
         for block in blocks:
-            term = block.ops[-1]
-            if isinstance(term, Goto):
-                term.label = get_real_target(term.label)
-            elif isinstance(term, Branch):
-                term.true = get_real_target(term.true)
-                term.false = get_real_target(term.false)
+            for i, tgt in enumerate(block.terminator.targets()):
+                block.terminator.set_target(i, get_real_target(tgt))
 
         # Then delete any blocks that have no predecessors
         changed = False
@@ -123,154 +151,186 @@ def cleanup_cfg(blocks: List[BasicBlock]) -> None:
                 changed = True
 
 
-T = TypeVar('T')
+T = TypeVar("T")
 
 AnalysisDict = Dict[Tuple[BasicBlock, int], Set[T]]
 
 
 class AnalysisResult(Generic[T]):
-    def __init__(self, before: 'AnalysisDict[T]', after: 'AnalysisDict[T]') -> None:
+    def __init__(self, before: AnalysisDict[T], after: AnalysisDict[T]) -> None:
         self.before = before
         self.after = after
 
     def __str__(self) -> str:
-        return 'before: %s\nafter: %s\n' % (self.before, self.after)
+        return f"before: {self.before}\nafter: {self.after}\n"
 
 
-GenAndKill = Tuple[Set[Value], Set[Value]]
+GenAndKill = Tuple[Set[T], Set[T]]
 
 
-class BaseAnalysisVisitor(OpVisitor[GenAndKill]):
-    def visit_goto(self, op: Goto) -> GenAndKill:
+class BaseAnalysisVisitor(OpVisitor[GenAndKill[T]]):
+    def visit_goto(self, op: Goto) -> GenAndKill[T]:
         return set(), set()
 
     @abstractmethod
-    def visit_register_op(self, op: RegisterOp) -> GenAndKill:
+    def visit_register_op(self, op: RegisterOp) -> GenAndKill[T]:
         raise NotImplementedError
 
     @abstractmethod
-    def visit_assign(self, op: Assign) -> GenAndKill:
+    def visit_assign(self, op: Assign) -> GenAndKill[T]:
         raise NotImplementedError
 
     @abstractmethod
-    def visit_set_mem(self, op: SetMem) -> GenAndKill:
+    def visit_assign_multi(self, op: AssignMulti) -> GenAndKill[T]:
         raise NotImplementedError
 
-    def visit_call(self, op: Call) -> GenAndKill:
+    @abstractmethod
+    def visit_set_mem(self, op: SetMem) -> GenAndKill[T]:
+        raise NotImplementedError
+
+    def visit_call(self, op: Call) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_method_call(self, op: MethodCall) -> GenAndKill:
+    def visit_method_call(self, op: MethodCall) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_load_error_value(self, op: LoadErrorValue) -> GenAndKill:
+    def visit_load_error_value(self, op: LoadErrorValue) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_get_attr(self, op: GetAttr) -> GenAndKill:
+    def visit_load_literal(self, op: LoadLiteral) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_set_attr(self, op: SetAttr) -> GenAndKill:
+    def visit_get_attr(self, op: GetAttr) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_load_static(self, op: LoadStatic) -> GenAndKill:
+    def visit_set_attr(self, op: SetAttr) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_init_static(self, op: InitStatic) -> GenAndKill:
+    def visit_load_static(self, op: LoadStatic) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_tuple_get(self, op: TupleGet) -> GenAndKill:
+    def visit_init_static(self, op: InitStatic) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_tuple_set(self, op: TupleSet) -> GenAndKill:
+    def visit_tuple_get(self, op: TupleGet) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_box(self, op: Box) -> GenAndKill:
+    def visit_tuple_set(self, op: TupleSet) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_unbox(self, op: Unbox) -> GenAndKill:
+    def visit_box(self, op: Box) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_cast(self, op: Cast) -> GenAndKill:
+    def visit_unbox(self, op: Unbox) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_raise_standard_error(self, op: RaiseStandardError) -> GenAndKill:
+    def visit_cast(self, op: Cast) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_call_c(self, op: CallC) -> GenAndKill:
+    def visit_raise_standard_error(self, op: RaiseStandardError) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_truncate(self, op: Truncate) -> GenAndKill:
+    def visit_call_c(self, op: CallC) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_load_global(self, op: LoadGlobal) -> GenAndKill:
+    def visit_truncate(self, op: Truncate) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_int_op(self, op: IntOp) -> GenAndKill:
+    def visit_extend(self, op: Extend) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_comparison_op(self, op: ComparisonOp) -> GenAndKill:
+    def visit_load_global(self, op: LoadGlobal) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_load_mem(self, op: LoadMem) -> GenAndKill:
+    def visit_int_op(self, op: IntOp) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_get_element_ptr(self, op: GetElementPtr) -> GenAndKill:
+    def visit_comparison_op(self, op: ComparisonOp) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
-    def visit_load_address(self, op: LoadAddress) -> GenAndKill:
+    def visit_load_mem(self, op: LoadMem) -> GenAndKill[T]:
+        return self.visit_register_op(op)
+
+    def visit_get_element_ptr(self, op: GetElementPtr) -> GenAndKill[T]:
+        return self.visit_register_op(op)
+
+    def visit_load_address(self, op: LoadAddress) -> GenAndKill[T]:
+        return self.visit_register_op(op)
+
+    def visit_keep_alive(self, op: KeepAlive) -> GenAndKill[T]:
         return self.visit_register_op(op)
 
 
-class DefinedVisitor(BaseAnalysisVisitor):
+class DefinedVisitor(BaseAnalysisVisitor[Value]):
     """Visitor for finding defined registers.
 
     Note that this only deals with registers and not temporaries, on
     the assumption that we never access temporaries when they might be
     undefined.
+
+    If strict_errors is True, then we regard any use of LoadErrorValue
+    as making a register undefined. Otherwise we only do if
+    `undefines` is set on the error value.
+
+    This lets us only consider the things we care about during
+    uninitialized variable checking while capturing all possibly
+    undefined things for refcounting.
     """
 
-    def visit_branch(self, op: Branch) -> GenAndKill:
+    def __init__(self, strict_errors: bool = False) -> None:
+        self.strict_errors = strict_errors
+
+    def visit_branch(self, op: Branch) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_return(self, op: Return) -> GenAndKill:
+    def visit_return(self, op: Return) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_unreachable(self, op: Unreachable) -> GenAndKill:
+    def visit_unreachable(self, op: Unreachable) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_register_op(self, op: RegisterOp) -> GenAndKill:
+    def visit_register_op(self, op: RegisterOp) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_assign(self, op: Assign) -> GenAndKill:
+    def visit_assign(self, op: Assign) -> GenAndKill[Value]:
         # Loading an error value may undefine the register.
-        if isinstance(op.src, LoadErrorValue) and op.src.undefines:
+        if isinstance(op.src, LoadErrorValue) and (op.src.undefines or self.strict_errors):
             return set(), {op.dest}
         else:
             return {op.dest}, set()
 
-    def visit_set_mem(self, op: SetMem) -> GenAndKill:
+    def visit_assign_multi(self, op: AssignMulti) -> GenAndKill[Value]:
+        # Array registers are special and we don't track the definedness of them.
+        return set(), set()
+
+    def visit_set_mem(self, op: SetMem) -> GenAndKill[Value]:
         return set(), set()
 
 
-def analyze_maybe_defined_regs(blocks: List[BasicBlock],
-                               cfg: CFG,
-                               initial_defined: Set[Value]) -> AnalysisResult[Value]:
+def analyze_maybe_defined_regs(
+    blocks: list[BasicBlock], cfg: CFG, initial_defined: set[Value]
+) -> AnalysisResult[Value]:
     """Calculate potentially defined registers at each CFG location.
 
     A register is defined if it has a value along some path from the initial location.
     """
-    return run_analysis(blocks=blocks,
-                        cfg=cfg,
-                        gen_and_kill=DefinedVisitor(),
-                        initial=initial_defined,
-                        backward=False,
-                        kind=MAYBE_ANALYSIS)
+    return run_analysis(
+        blocks=blocks,
+        cfg=cfg,
+        gen_and_kill=DefinedVisitor(),
+        initial=initial_defined,
+        backward=False,
+        kind=MAYBE_ANALYSIS,
+    )
 
 
 def analyze_must_defined_regs(
-        blocks: List[BasicBlock],
-        cfg: CFG,
-        initial_defined: Set[Value],
-        regs: Iterable[Value]) -> AnalysisResult[Value]:
+    blocks: list[BasicBlock],
+    cfg: CFG,
+    initial_defined: set[Value],
+    regs: Iterable[Value],
+    strict_errors: bool = False,
+) -> AnalysisResult[Value]:
     """Calculate always defined registers at each CFG location.
 
     This analysis can work before exception insertion, since it is a
@@ -280,80 +340,89 @@ def analyze_must_defined_regs(
     A register is defined if it has a value along all paths from the
     initial location.
     """
-    return run_analysis(blocks=blocks,
-                        cfg=cfg,
-                        gen_and_kill=DefinedVisitor(),
-                        initial=initial_defined,
-                        backward=False,
-                        kind=MUST_ANALYSIS,
-                        universe=set(regs))
+    return run_analysis(
+        blocks=blocks,
+        cfg=cfg,
+        gen_and_kill=DefinedVisitor(strict_errors=strict_errors),
+        initial=initial_defined,
+        backward=False,
+        kind=MUST_ANALYSIS,
+        universe=set(regs),
+    )
 
 
-class BorrowedArgumentsVisitor(BaseAnalysisVisitor):
-    def __init__(self, args: Set[Value]) -> None:
+class BorrowedArgumentsVisitor(BaseAnalysisVisitor[Value]):
+    def __init__(self, args: set[Value]) -> None:
         self.args = args
 
-    def visit_branch(self, op: Branch) -> GenAndKill:
+    def visit_branch(self, op: Branch) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_return(self, op: Return) -> GenAndKill:
+    def visit_return(self, op: Return) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_unreachable(self, op: Unreachable) -> GenAndKill:
+    def visit_unreachable(self, op: Unreachable) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_register_op(self, op: RegisterOp) -> GenAndKill:
+    def visit_register_op(self, op: RegisterOp) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_assign(self, op: Assign) -> GenAndKill:
+    def visit_assign(self, op: Assign) -> GenAndKill[Value]:
         if op.dest in self.args:
             return set(), {op.dest}
         return set(), set()
 
-    def visit_set_mem(self, op: SetMem) -> GenAndKill:
+    def visit_assign_multi(self, op: AssignMulti) -> GenAndKill[Value]:
+        return set(), set()
+
+    def visit_set_mem(self, op: SetMem) -> GenAndKill[Value]:
         return set(), set()
 
 
 def analyze_borrowed_arguments(
-        blocks: List[BasicBlock],
-        cfg: CFG,
-        borrowed: Set[Value]) -> AnalysisResult[Value]:
+    blocks: list[BasicBlock], cfg: CFG, borrowed: set[Value]
+) -> AnalysisResult[Value]:
     """Calculate arguments that can use references borrowed from the caller.
 
     When assigning to an argument, it no longer is borrowed.
     """
-    return run_analysis(blocks=blocks,
-                        cfg=cfg,
-                        gen_and_kill=BorrowedArgumentsVisitor(borrowed),
-                        initial=borrowed,
-                        backward=False,
-                        kind=MUST_ANALYSIS,
-                        universe=borrowed)
+    return run_analysis(
+        blocks=blocks,
+        cfg=cfg,
+        gen_and_kill=BorrowedArgumentsVisitor(borrowed),
+        initial=borrowed,
+        backward=False,
+        kind=MUST_ANALYSIS,
+        universe=borrowed,
+    )
 
 
-class UndefinedVisitor(BaseAnalysisVisitor):
-    def visit_branch(self, op: Branch) -> GenAndKill:
+class UndefinedVisitor(BaseAnalysisVisitor[Value]):
+    def visit_branch(self, op: Branch) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_return(self, op: Return) -> GenAndKill:
+    def visit_return(self, op: Return) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_unreachable(self, op: Unreachable) -> GenAndKill:
+    def visit_unreachable(self, op: Unreachable) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_register_op(self, op: RegisterOp) -> GenAndKill:
+    def visit_register_op(self, op: RegisterOp) -> GenAndKill[Value]:
         return set(), {op} if not op.is_void else set()
 
-    def visit_assign(self, op: Assign) -> GenAndKill:
+    def visit_assign(self, op: Assign) -> GenAndKill[Value]:
         return set(), {op.dest}
 
-    def visit_set_mem(self, op: SetMem) -> GenAndKill:
+    def visit_assign_multi(self, op: AssignMulti) -> GenAndKill[Value]:
+        return set(), {op.dest}
+
+    def visit_set_mem(self, op: SetMem) -> GenAndKill[Value]:
         return set(), set()
 
 
-def analyze_undefined_regs(blocks: List[BasicBlock],
-                           cfg: CFG,
-                           initial_defined: Set[Value]) -> AnalysisResult[Value]:
+def analyze_undefined_regs(
+    blocks: list[BasicBlock], cfg: CFG, initial_defined: set[Value]
+) -> AnalysisResult[Value]:
     """Calculate potentially undefined registers at each CFG location.
 
     A register is undefined if there is some path from initial block
@@ -362,15 +431,17 @@ def analyze_undefined_regs(blocks: List[BasicBlock],
     Function arguments are assumed to be always defined.
     """
     initial_undefined = set(all_values([], blocks)) - initial_defined
-    return run_analysis(blocks=blocks,
-                        cfg=cfg,
-                        gen_and_kill=UndefinedVisitor(),
-                        initial=initial_undefined,
-                        backward=False,
-                        kind=MAYBE_ANALYSIS)
+    return run_analysis(
+        blocks=blocks,
+        cfg=cfg,
+        gen_and_kill=UndefinedVisitor(),
+        initial=initial_undefined,
+        backward=False,
+        kind=MAYBE_ANALYSIS,
+    )
 
 
-def non_trivial_sources(op: Op) -> Set[Value]:
+def non_trivial_sources(op: Op) -> set[Value]:
     result = set()
     for source in op.sources():
         if not isinstance(source, Integer):
@@ -378,46 +449,50 @@ def non_trivial_sources(op: Op) -> Set[Value]:
     return result
 
 
-class LivenessVisitor(BaseAnalysisVisitor):
-    def visit_branch(self, op: Branch) -> GenAndKill:
+class LivenessVisitor(BaseAnalysisVisitor[Value]):
+    def visit_branch(self, op: Branch) -> GenAndKill[Value]:
         return non_trivial_sources(op), set()
 
-    def visit_return(self, op: Return) -> GenAndKill:
+    def visit_return(self, op: Return) -> GenAndKill[Value]:
         if not isinstance(op.value, Integer):
             return {op.value}, set()
         else:
             return set(), set()
 
-    def visit_unreachable(self, op: Unreachable) -> GenAndKill:
+    def visit_unreachable(self, op: Unreachable) -> GenAndKill[Value]:
         return set(), set()
 
-    def visit_register_op(self, op: RegisterOp) -> GenAndKill:
+    def visit_register_op(self, op: RegisterOp) -> GenAndKill[Value]:
         gen = non_trivial_sources(op)
         if not op.is_void:
             return gen, {op}
         else:
             return gen, set()
 
-    def visit_assign(self, op: Assign) -> GenAndKill:
+    def visit_assign(self, op: Assign) -> GenAndKill[Value]:
         return non_trivial_sources(op), {op.dest}
 
-    def visit_set_mem(self, op: SetMem) -> GenAndKill:
+    def visit_assign_multi(self, op: AssignMulti) -> GenAndKill[Value]:
+        return non_trivial_sources(op), {op.dest}
+
+    def visit_set_mem(self, op: SetMem) -> GenAndKill[Value]:
         return non_trivial_sources(op), set()
 
 
-def analyze_live_regs(blocks: List[BasicBlock],
-                      cfg: CFG) -> AnalysisResult[Value]:
+def analyze_live_regs(blocks: list[BasicBlock], cfg: CFG) -> AnalysisResult[Value]:
     """Calculate live registers at each CFG location.
 
     A register is live at a location if it can be read along some CFG path starting
     from the location.
     """
-    return run_analysis(blocks=blocks,
-                        cfg=cfg,
-                        gen_and_kill=LivenessVisitor(),
-                        initial=set(),
-                        backward=True,
-                        kind=MAYBE_ANALYSIS)
+    return run_analysis(
+        blocks=blocks,
+        cfg=cfg,
+        gen_and_kill=LivenessVisitor(),
+        initial=set(),
+        backward=True,
+        kind=MAYBE_ANALYSIS,
+    )
 
 
 # Analysis kinds
@@ -425,16 +500,15 @@ MUST_ANALYSIS = 0
 MAYBE_ANALYSIS = 1
 
 
-# TODO the return type of this function is too complicated. Abtract it into its
-# own class.
-
-def run_analysis(blocks: List[BasicBlock],
-                 cfg: CFG,
-                 gen_and_kill: OpVisitor[Tuple[Set[T], Set[T]]],
-                 initial: Set[T],
-                 kind: int,
-                 backward: bool,
-                 universe: Optional[Set[T]] = None) -> AnalysisResult[T]:
+def run_analysis(
+    blocks: list[BasicBlock],
+    cfg: CFG,
+    gen_and_kill: OpVisitor[GenAndKill[T]],
+    initial: set[T],
+    kind: int,
+    backward: bool,
+    universe: set[T] | None = None,
+) -> AnalysisResult[T]:
     """Run a general set-based data flow analysis.
 
     Args:
@@ -457,15 +531,15 @@ def run_analysis(blocks: List[BasicBlock],
 
     # Calculate kill and gen sets for entire basic blocks.
     for block in blocks:
-        gen = set()  # type: Set[T]
-        kill = set()  # type: Set[T]
+        gen: set[T] = set()
+        kill: set[T] = set()
         ops = block.ops
         if backward:
             ops = list(reversed(ops))
         for op in ops:
             opgen, opkill = op.accept(gen_and_kill)
-            gen = ((gen - opkill) | opgen)
-            kill = ((kill - opgen) | opkill)
+            gen = (gen - opkill) | opgen
+            kill = (kill - opgen) | opkill
         block_gen[block] = gen
         block_kill[block] = kill
 
@@ -474,8 +548,8 @@ def run_analysis(blocks: List[BasicBlock],
     if not backward:
         worklist = worklist[::-1]  # Reverse for a small performance improvement
     workset = set(worklist)
-    before = {}  # type: Dict[BasicBlock, Set[T]]
-    after = {}  # type: Dict[BasicBlock, Set[T]]
+    before: dict[BasicBlock, set[T]] = {}
+    after: dict[BasicBlock, set[T]] = {}
     for block in blocks:
         if kind == MAYBE_ANALYSIS:
             before[block] = set()
@@ -497,7 +571,7 @@ def run_analysis(blocks: List[BasicBlock],
         label = worklist.pop()
         workset.remove(label)
         if pred_map[label]:
-            new_before = None  # type: Union[Set[T], None]
+            new_before: set[T] | None = None
             for pred in pred_map[label]:
                 if new_before is None:
                     new_before = set(after[pred])
@@ -518,12 +592,12 @@ def run_analysis(blocks: List[BasicBlock],
         after[label] = new_after
 
     # Run algorithm for each basic block to generate opcode-level sets.
-    op_before = {}  # type: Dict[Tuple[BasicBlock, int], Set[T]]
-    op_after = {}  # type: Dict[Tuple[BasicBlock, int], Set[T]]
+    op_before: dict[tuple[BasicBlock, int], set[T]] = {}
+    op_after: dict[tuple[BasicBlock, int], set[T]] = {}
     for block in blocks:
         label = block
         cur = before[label]
-        ops_enum = enumerate(block.ops)  # type: Iterator[Tuple[int, Op]]
+        ops_enum: Iterator[tuple[int, Op]] = enumerate(block.ops)
         if backward:
             ops_enum = reversed(list(ops_enum))
         for idx, op in ops_enum:
