@@ -1,4 +1,4 @@
-from typing import List
+from __future__ import annotations
 
 import mypy.errors
 
@@ -7,20 +7,20 @@ class Errors:
     def __init__(self) -> None:
         self.num_errors = 0
         self.num_warnings = 0
-        self._errors = mypy.errors.Errors()
+        self._errors = mypy.errors.Errors(hide_error_codes=True)
 
     def error(self, msg: str, path: str, line: int) -> None:
-        self._errors.report(line, None, msg, severity='error', file=path)
+        self._errors.report(line, None, msg, severity="error", file=path)
         self.num_errors += 1
 
     def note(self, msg: str, path: str, line: int) -> None:
-        self._errors.report(line, None, msg, severity='note', file=path)
+        self._errors.report(line, None, msg, severity="note", file=path)
 
     def warning(self, msg: str, path: str, line: int) -> None:
-        self._errors.report(line, None, msg, severity='warning', file=path)
+        self._errors.report(line, None, msg, severity="warning", file=path)
         self.num_warnings += 1
 
-    def new_messages(self) -> List[str]:
+    def new_messages(self) -> list[str]:
         return self._errors.new_messages()
 
     def flush_errors(self) -> None:
