@@ -4,7 +4,7 @@ class MyPlugin(Plugin):
     def get_function_hook(self, fullname):
         if fullname == '__main__.f':
             return my_hook
-        assert fullname is not None
+        assert fullname
         return None
 
 def my_hook(ctx):
