@@ -1,4 +1,4 @@
-from typing import Optional, Callable, List, Tuple
+from typing import List, Tuple
 
 from mypy.plugin import Plugin
 from mypy.nodes import MypyFile
@@ -6,7 +6,7 @@ from mypy.nodes import MypyFile
 
 class DepsPlugin(Plugin):
     def get_additional_deps(self, file: MypyFile) -> List[Tuple[int, str, int]]:
-        if file.fullname() == '__main__':
+        if file.fullname == '__main__':
             return [(10, 'err', -1)]
         return []
 
